/*
 *  NS-Link support daemon for Pepperl+Fuchs DeviceMaster products.
 *
 *  Copyright (C) Pepperl+Fuchs.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <poll.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <syslog.h>
#include <grp.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netpacket/packet.h>
#include <net/if.h>
#include <netdb.h>
#include <linux/kdev_t.h>
#ifdef _USE_OPENSSL
#include <sys/un.h>
#include <pthread.h>
#include <openssl/opensslv.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/bio.h>
#endif

#include "nslink.h"
#include "nslinkd.h"
#include "version.h"
//#include "connectsecure.h"

const char *not_loaded = "NS-Link driver not loaded?";
#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)
const char *pid_file = TOSTRING(PIDPATH) "/nslinkd.pid";

static pid_t child_pids[MAX_NRP_SIS];
static pid_t parent_pid;
static int child_socket;
static int driver_fd;

static char *mactoa(unsigned char *m);

int format_inet_addr(char *buf, unsigned maxlen, struct sockaddr_storage *ss)
{
  if ((ss->ss_family) == AF_INET)
    inet_ntop(AF_INET, &((struct sockaddr_in*)ss)->sin_addr, buf, maxlen-1);
  else
    inet_ntop(AF_INET6, &((struct sockaddr_in6*)ss)->sin6_addr, buf, maxlen-1);
  return strlen(buf);
}  

const char *inet_addr_string(struct sockaddr_storage *ss)
{
  static char buf[128];
  format_inet_addr(buf, sizeof buf, ss);
  return buf;
}

static int get_major_number(const char *name)
{
        FILE *f;
        char buf[80], devname[80];
        int n, major = -1;

        f = fopen("/proc/devices", "r");
        if (!f)
                return -1;
        while (!feof(f)) {
                if (!fgets(buf, sizeof(buf), f))
                        break;
                if (sscanf(buf, "%d %s", &n, devname) != 2)
                        continue;
                if (!strcmp(devname, name)) {
                        major = n;
                        break;
                }
        }
        fclose(f);
        return major;
}

// return a random, unused temporary device name

static char *myTempName(void)
{
        int tries;
        static char path[256];
        struct stat statbuf;

        srandom(time(NULL));
        for (tries=0; tries<10; ++tries) {
                sprintf(path,"/dev/nslink.%08lx.%08lx",(long)time(NULL),(long)random());
                if (stat(path,&statbuf))
                        return path;
        }
        return NULL;
}

static int open_dev(dev_t dev, int mode)
{
        char *fn;
        int fd;

        if ((fn = myTempName()) == NULL) {
                fprintf(stderr,"myTempName() failed\n");
                return -1;
        }
        if (mknod(fn, S_IFCHR | 0600, dev) != 0) {
                fprintf(stderr, "mknod(\"%s\", %04x, %lu) failed\n", fn, S_IFCHR | 0600, (unsigned long)dev);
                perror("mknod");
                return -1;
        }
        fd = open(fn, mode);
        if (fd < 0) {
                fprintf(stderr,"open(\"%s\",%d) failed\n",fn,mode);
                perror("open");
        }
        unlink(fn);
        return fd;
}

static int config_si(int fd,   struct daemon_si_box_config *config, int boxid)
{
        struct si_box_config cfg;
        int ret, i;

        //syslog(LOG_INFO, "config_si(fd=%d, boxid=%d, ...)\n",fd,boxid);  /* debug */

        memset(&cfg, 0, sizeof cfg);
        cfg.boxid = boxid;
        cfg.base_port = config->base_port;
        cfg.num_ports = config->num_lines;
        memcpy(cfg.ether_addr, config->ether_addr, 6);
        strcpy(cfg.interface, config->interface);
        cfg.ip_addr = config->ip_addr;
        cfg.link_timeout = config->link_timeout;
        cfg.scan_period = config->scan_period;
        cfg.rexmit_timer = config->rexmit_timer;
        cfg.arch_id = config->arch_id;
        cfg.ModelId = config->ModelId;
        for (i=0; i<MAX_NRP_PORTS; ++i) {
                cfg.rsmode[i] = config->rsmode[i];
                cfg.low_latency[i] = config->low_latency[i];
                cfg.inter_char_timeout[i] = config->inter_char_timeout[i];
                cfg.rx_fifo_disable[i] = config->rx_fifo_disable[i];
                cfg.term_res_tx_en[i] = config->term_res_tx_en[i];
                cfg.term_res_rx_en[i] = config->term_res_rx_en[i];
        }
        //syslog(LOG_INFO,"boxid %d link timeout = %d seconds\n",boxid,config->link_timeout);  /* debug */
        ret = ioctl(fd, SI_SET_BOX_CONFIG, &cfg);
        if (ret) {
                syslog(LOG_ERR, "Couldn't config NS-Link box #%d (errno %d)",boxid,errno);
                return errno;
        }
        //syslog(LOG_INFO,"config_si return\n");  /* debug */
        return 0;
}


static int become_daemon(int keep_fd)
{
        int child_pid, fd, max_files;

        child_pid = fork();
	syslog(LOG_INFO, "Fork become_daemon %u (%u)\n", child_pid, getpid());
        if (child_pid == -1)
                return -1;
        if (child_pid) {
		sleep(1);  // allow time for child to create PID file before we exit
		exit(0);
	}
        (void) chdir("/");
        fd = open("/dev/null", O_RDWR, 0);
        if (fd >= 0) {
                (void) dup2(fd, 0);
                (void) dup2(fd, 1);
                (void) dup2(fd, 2);
                if (fd > 2)
                        (void) close(fd);
        }

        max_files = sysconf(_SC_OPEN_MAX);
        for (fd = 3; fd < max_files; ++fd)
                if (fd != keep_fd)
                        (void) close(fd);

        return (0);
}

static void check_pid_file()
{
        FILE *f;
        int pid;

        if (access(pid_file, R_OK) == 0) {
                f = fopen(pid_file, "r");
                if (f) {
                        pid = 0;
                        fscanf(f, "%d", &pid);
                        if (pid && !kill(pid, 0)) {
                                fprintf(stderr, "NS-Link daemon already running!\n");
                                syslog(LOG_ERR, "NS-Link daemon already running");
                                exit(1);
                        }
                        fclose(f);
                }
        }
}

static void put_pid_file()
{
        FILE *f;

        f = fopen(pid_file, "w");
        if (!f) {
                syslog(LOG_WARNING, "Warning: couldn't write pid file %s", pid_file);
                return;
        }
        parent_pid = getpid();
        fprintf(f, "%d\n", parent_pid);
        fclose(f);
}

static void create_device(const char *prefix, int major, int num, gid_t def_group)
{
        char name[80];
        struct stat st;
        dev_t device;
        mode_t mode = 0620;  // default file protection
        uid_t  owner = 0;    // default owner
        gid_t  group = def_group;

        sprintf(name, "%s%d", prefix, num);
        // syslog(LOG_INFO, "Device created %s\n", name);
        device = MKDEV(major, num);
        if (lstat(name, &st) == 0) {
                if ((st.st_rdev == device) && S_ISCHR(st.st_mode))
                        return;
#if 0

                syslog(LOG_INFO, "%s has dev %d, mode %o wanted %d",
                       name, st.st_rdev, st.st_mode, device);
#endif
                // device exists, but doesn't have right major/minor numbers,
                // so remember owner/group/mode so we can create new device
                // with same properties as old one except w/ right major/minor.

                mode = st.st_mode & 0777;
                owner = st.st_uid;
                group = st.st_gid;

                (void) unlink(name);
        }

        umask(0);

        if (mknod(name, S_IFCHR | mode, device) < 0) {
                perror(name);
                syslog(LOG_ERR, "Couldn't create device %s: %m", name);
        }
        if (chown(name, owner, group) < 0) {
                perror(name);
                syslog(LOG_ERR, "Couldn't chmod device %s: %m", name);
        }
}

/*
 * This function creates all of the devices
 */
static void create_devices(int ctlfd)
{
        int i, num_ports;
        struct group *grp;
        gid_t tty_group = 0;
        int major, callout_major;
        tBoxPortCount bp;

        num_ports = 0;
        for (i = 0; i < num_box_config; i++)
                num_ports += box_config[i].num_lines;

        syslog(LOG_INFO, "Registering %d devices, %d ports\n", num_box_config, num_ports);

        bp.boxCount = num_box_config;
        bp.portCount = num_ports;

        if (ioctl(ctlfd,SI_REGISTER_PORTS,&bp))
                syslog(LOG_ERR, "Couldn't register ports\n");

        syslog(LOG_INFO, "Creating devices\n");

        major = get_major_number("ttySI");
        callout_major = get_major_number("cuSI");

        if (major <= 0 && callout_major <= 0)
                return;

        grp = getgrnam("tty");
        if (grp)
                tty_group = grp->gr_gid;

        return;
        for (i=0; i<num_ports; ++i) {
                if (major > 0)
                        create_device("/dev/ttySI", major, i, tty_group);
                if (callout_major > 0)
                        create_device("/dev/cuSI", callout_major, i, tty_group);
        }
}


static void remove_devices(void)
{
        int i, num_ports;

        num_ports = 0;
        for (i = 0; i < num_box_config; i++)
                num_ports += box_config[i].num_lines;

        syslog(LOG_INFO, "Removing %d devices\n", num_ports);

        for (i=0; i<num_ports; ++i) {
          char s[32];
          snprintf(s, sizeof s, "/dev/ttySI%d", i);
          unlink(s);
          snprintf(s, sizeof s, "/dev/ttycuSI%d", i);
          unlink(s);
        }
}


#ifdef _USE_OPENSSL
// Control the service thread.
int stopService = 0;
int serviceRunning = 0;

pthread_mutex_t stopMutex;

static int getStopIssued(void) {
  int ret = 0;
  pthread_mutex_lock(&stopMutex);
  ret = stopService;
  pthread_mutex_unlock(&stopMutex);
  return ret;
}

static void setStopService(int val) {
  pthread_mutex_lock(&stopMutex);
  stopService = val;
  pthread_mutex_unlock(&stopMutex);
}

static int getServiceRunning(void) {
  return serviceRunning;
}

static void doStopService(char *phase)
{
  // Stop the service thread if it is running.
  if (getServiceRunning()) {
    syslog(LOG_INFO, "Stop service %s (%u)\n", phase, getpid());
    setStopService(1);
    while (getServiceRunning()) {
      setStopService(1);
      syslog(LOG_INFO, "Waiting service stop (%s) running %d...(%u)\n",
	     phase, serviceRunning, getpid());
      sleep(1);
    }
  }
}

#endif // _USE_OPENSSL

#define HOLD_OPEN 1

static void sigterm_handler(int signum)
{
  int i;

#ifdef NSLINKCTL
  if (0) {
    // Try to just stay open if there is a port open. Mantis_1013 MNSWELP-120.
    // This won't work. The system will send a SIGKILL if sigterm doesn't work.
    // See the nslinkctl branch for a scheme that does work. Control driver with a python script.
    struct si_driver_state dstate;
    int rc;
    syslog(LOG_INFO, "fd %d SIG(%d) is received (%u)\n", driver_fd, signum, getpid());
    syslog(LOG_INFO, "SIGTERM %d SIGKILL %d SIGHUP %d SIGINT %d\n",
	   SIGTERM, SIGKILL, SIGHUP, SIGINT);
    rc = ioctl(driver_fd, SI_GET_DRIVER_STATE, &dstate);
    if (rc) {
      syslog(LOG_ERR, "Could not get driver state %d\n", rc);
      exit(1);
    }
    syslog(LOG_INFO, "Driver state total_count %u (%u)\n", dstate.total_count, getpid());
    if (dstate.total_count && HOLD_OPEN) {
      syslog(LOG_INFO, "Driver state total_count %u SIGTERM ignored! (%u)\n", dstate.total_count, getpid());
      return;
    }
  }
#endif
  syslog(LOG_INFO, "Received SIGTERM/SIGINT (%u)\n", getpid());
	        
        if (parent_pid == getpid()) {
	        // systemctl seems to send it's own sigterm to the children. Let that take effect before
	        // we do it again.
	        sleep(2);
                for (i = 0; i < num_box_config; i++)
                        if (child_pids[i])
			  {
			    int wstatus;
			    pid_t pid = child_pids[i];
			    syslog(LOG_INFO, "Send sigterm to %u (%u)\n", pid, getpid());
			    kill(pid, SIGTERM);
			    waitpid(pid, &wstatus, 0);
			    //syslog(LOG_INFO, "Wait complete %u (%u)\n", pid, getpid());
			  }
                syslog(LOG_INFO, "Exiting");
                remove_devices();
                (void) unlink(pid_file);
        } else {
#ifdef _USE_OPENSSL
	  // Stop the service thread if it is running.
	  doStopService("sigterm");
#endif // _USE_OPENSSL
          close(child_socket);
        }
	syslog(LOG_INFO,"sigterm child (%u) exit\n", getpid());
        exit(0);
}

static void sighup_handler()
{
        syslog(LOG_INFO, "sighup (%u)\n", getpid());
        signal(SIGHUP, sighup_handler);
        return;
}

static int packetReadTimeout(int socket, unsigned char *macAddr, unsigned char index, void *b, int count,  unsigned usTimeout)
{
        unsigned char *buf = b;
        fd_set r,e;
        struct timeval tv;
        int s;
        int n;


        FD_ZERO(&r);
        FD_ZERO(&e);
        FD_SET(socket,&r);
        FD_SET(socket,&e);
        tv.tv_sec = usTimeout/1000000;
        tv.tv_usec = usTimeout % 1000000;

        while (1) {
                s = select(socket+1,&r,NULL,&e,&tv);

                if (s==0)
                        return 0;

                if (!FD_ISSET(socket,&r))
                        return 0;

                n = read(socket,buf,count);

                if (memcmp(buf+6,macAddr,6))
                        continue;  // not from the right box: count on fact that tv is time remaining

                if (buf[15] != index)
                        continue;  // not intended for daemon

                return n;
        }
}

#define READ_TIMEOUT 2000

static int read_with_timeout(int fd, void *buf, size_t count, unsigned timeout_ms)
{
        struct pollfd pfd = { fd, POLLIN, 0 };

        int n = poll(&pfd, 1, timeout_ms);

        if (n < 0)
                return n;

        if (n == 0) {
                errno = ETIME;
                return -1;
        }

        if (pfd.revents & POLLIN)
                return read(fd,buf,count);
        else
                return 0;
}

int parse_host(const char *hostname, struct sockaddr_storage *addr)
{
        int r;
        struct addrinfo *res;
        struct addrinfo hints = { .ai_socktype = SOCK_STREAM, .ai_family   = AF_UNSPEC};

        memset(addr, 0, sizeof *addr);

        r = getaddrinfo(hostname,"4606",&hints,&res);

        if (r || !res) {
                syslog(LOG_ERR,"Can't resolve host/address '%s': %s\n",hostname,gai_strerror(r));
                return -1;
        }

        if (res->ai_addrlen > sizeof *addr) {
                syslog(LOG_ERR,"Returned address info for '%s' too large for storage\n",hostname);
                freeaddrinfo(res);
                return -1;
        }

        memcpy(addr,res->ai_addr,res->ai_addrlen);

        if (res->ai_next)
                syslog(LOG_WARNING,"Host '%s' resolved to multiple IP addresses. Using %s\n",hostname,inet_addr_string((void*)(res->ai_addr)));

        return 0;
}

#define msleep(ms) usleep((ms)*1000)

// Dump packet contents details.
//#define PACKET_DEBUG
//#define SOCKET_TEST
//#define UDP_SENDER
//#define COUNT_LOOPS

#ifdef UDP_SENDER
// Some stuff to send diagnostic UDP packets. Wireshark udp.port == 10000

int sendUDP(int port, char *message)
{
  int result = 0;
  struct sockaddr_in server_addr;
  int server_struct_length = sizeof(server_addr);

  // My IP address.
  char *IP_address = "192.168.250.7";

  // Set port and IP:
  server_addr.sin_family = AF_INET;
  server_addr.sin_port = htons(port);
  server_addr.sin_addr.s_addr = inet_addr(IP_address);
  
  int sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  // Bind to the set port and IP:
  if(bind(sock, (struct sockaddr*)&server_addr, sizeof(server_addr)) < 0){
        syslog(LOG_INFO, "Couldn't bind to the port\n");
        return -1;
  }
  // Send the message
  // Set port and IP:
  server_addr.sin_family = AF_INET;
  server_addr.sin_port = htons(port);
  server_addr.sin_addr.s_addr = inet_addr("192.168.250.202");
  
  result = sendto(sock, message, strlen(message), 0,
		  (struct sockaddr*)&server_addr, server_struct_length);
  if (result < 0) {
    syslog(LOG_INFO, "Unable to send message\n");
    return -1;
    }

  syslog(LOG_INFO, "sendUDP %d bytes sent\n", result);     
    
  return 0;
}

#endif

#ifdef PACKET_DEBUG
static void hexline(int offset, uint8_t *data, int size)
{
  char inbuf[100];
  char buf[256];
  int i;

  buf[0] = 0;
  sprintf(inbuf, "%03x: ", offset);
  strcat(buf, inbuf);
  for (i=0; i<size; i++) {
    sprintf(inbuf, "%02x ", data[i]);
    //sprintf(inbuf, "%02x %c ", data[i], data[i]);
    strcat(buf, inbuf);
  }
  syslog(LOG_INFO, "%s\n", buf);
}

static void hexdump(const char *name, uint8_t *data, unsigned len)
{
  int i=0, num;
  int inc=16;
  
  if (name)
    syslog(LOG_INFO, "%s:\n",name);
  while (i<len) {
    num = len - i;
    if (num > inc) num=inc;
    hexline(i, &(data[i]), num);
    i += num;
  }
}

#endif

#ifdef SOCKET_TEST

int shutSock = -1;
pthread tidTest;
  
void *Tester(void *arg)
// Do something to test some things.
{
  sleep(5);
  if (shutSock < 0) return NULL;
  // Shutdown a socket to make sure we handle it correctly.
  syslog(LOG_INFO, "SOCKET SHUTDOWN!!!!!\n");
  shutdown(shutSock, SHUT_WR);
  return NULL;
}
#endif

// Are we running ssl?
int sslInited = 0;

// Try harder to report an error condition.
void on_error(char *prompt, int result)
{
  if (prompt != NULL) {
    syslog(LOG_ERR, "%s (%u)", prompt, getpid());
  }
  if (errno) {
    syslog(LOG_ERR, "errno %d '%s' (%u)", errno, strerror(errno), getpid());
  }
#ifdef _USE_OPENSSL
  if (sslInited) {
    BIO *bio = BIO_new(BIO_s_mem());
    ERR_print_errors(bio);
    char *buf = NULL;
    size_t len = BIO_get_mem_data(bio, &buf);
    if (len) {
      char *ret = (char *) calloc (1, 1 + len);
      if (ret) {
	memcpy (ret, buf, len);
	syslog(LOG_ERR, "%s (%u)", ret, getpid());
	free(ret);
      }
    }
    BIO_free (bio);
  }
#endif // _USE_OPENSSL
}

#ifdef _USE_OPENSSL

#define BUFSIZE 4096

struct secureConnection {
  int tcpSock;
  SSL_CTX *ctx;
  SSL *ssl;
  struct sockaddr_un unixAddr;
};

void *Service(void *arg)
// Run the connection between the unix socket and the tcp socket.
{
  fd_set active_fd_set, read_fd_set, write_fd_set;
  int fd;
  char addrstr[128];
  uint8_t tcpBuf[BUFSIZE];
  uint8_t  unixBuf[BUFSIZE];
  int unixSock, tcpSock;
  int tcpBytes, unixBytes;
  int bytes;
  int selectResult;
  struct secureConnection *conn;
  unsigned long long loopcount = 0;
  int maxfd;

  conn = (struct secureConnection *)arg;
  tcpSock = conn->tcpSock;
  strncpy(addrstr, (conn->unixAddr.sun_path)+1, sizeof(addrstr));
  syslog(LOG_INFO, "Unix socket name %s\n", addrstr);
#ifdef PACKET_DEBUG
  char tcpname[256];
  int tid;
  tid = (-1);
#ifdef SYS_gettid
  tid = syscall(SYS_gettid);
#endif
  sprintf(tcpname, "SSL TID %d %s", tid, addrstr);
  syslog(LOG_INFO, "Service tag %s\n", tcpname);
#endif

  if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == (-1)) {
    on_error("Thread socket creation error", 0);
    exit(-1);
  }
  if (bind(fd, (struct sockaddr *)&(conn->unixAddr), sizeof(struct sockaddr_un)) == -1) {
    on_error("Thread socket bind error", 0);
    exit(-1);
  }
  if (listen(fd, 5) == -1) {
    on_error("Unix server socket listen error", 0);
    exit(-1);
  }
  syslog(LOG_INFO, "Wait for connection...(%u)\n", getpid());
  if ((unixSock = accept(fd, NULL, NULL)) == -1) {
    on_error("Thread socket accept error", 0);
  }
  //syslog(LOG_INFO, "Connection accepted %d\n", unixSock);
  /* Initialize the set of active sockets. */
  //syslog(LOG_INFO, "Have unix socket %d\n", unixSock);
  // Set TCP socket non-blocking.
  syslog(LOG_INFO, "Set TCP socket non-blocking\n");
  fcntl(tcpSock, F_SETFL, fcntl(tcpSock, F_GETFL) | O_NONBLOCK);
  maxfd = 0;
  FD_ZERO(&active_fd_set);
  FD_SET(tcpSock, &active_fd_set);
  if (tcpSock > maxfd) maxfd = tcpSock;
  FD_SET(unixSock, &active_fd_set);
  if (unixSock > maxfd) maxfd = unixSock;
  tcpBytes = 0;
  unixBytes = 0;
  syslog(LOG_INFO, "Begin service loop unix socket %d tcp socket %d maxfd %d(%u)\n", unixSock, tcpSock,
	 maxfd, getpid());
  while (1) {
#ifdef COUNT_LOOPS
    if ((loopcount % 1) == 0) {
      syslog(LOG_INFO, "Loop count %lld\n", loopcount);
      sleep(1);
    }
#endif
    loopcount++;
    serviceRunning = 1;
    if (getStopIssued()) {
      syslog(LOG_INFO, "Stopping service loop (%u)\n", getpid());
      break;
    }
    // Read anytime from either socket.
    read_fd_set = active_fd_set;
    // Set wait for write ready when we have something to write.
    FD_ZERO(&write_fd_set);
    if (tcpBytes) {
      FD_SET(unixSock, &write_fd_set);
    }
    if (unixBytes) {
      FD_SET(tcpSock, &write_fd_set);
    }
    //write_fd_set = active_fd_set;
    // Check socket status
#ifdef PACKET_DEBUG
    // Report select status if we have pending write bytes.
    if (tcpBytes || unixBytes || 1) {
        syslog(LOG_INFO, "Call selekt r/w TCP %d %d/%d UNIX %d %d/%d (%u)\n",
	       unixBytes,
	       FD_ISSET(tcpSock, &read_fd_set),
	       FD_ISSET(tcpSock, &write_fd_set),
	       tcpBytes,
	       FD_ISSET(unixSock, &read_fd_set),
	       FD_ISSET(unixSock, &write_fd_set),
	       getpid());
    }
#endif
    if (0) {
      // Timeout for nonblocking behavior.
      struct timeval timeout;
      timeout.tv_sec  = 0;
      timeout.tv_usec = 100000;
      selectResult = select(maxfd + 1, &read_fd_set, &write_fd_set, NULL, &timeout);
    } else {
      // Timeout for nonblocking behavior.
      struct timespec timeout;
      sigset_t sigset;
      sigemptyset(&sigset);
      // Block SIGTERM?
      //sigaddset(&sigset, SIGTERM);
      timeout.tv_sec = 0;
      timeout.tv_nsec = 100000000;
      selectResult = pselect(maxfd + 1, &read_fd_set, &write_fd_set, NULL, &timeout, &sigset);
    }
#ifdef PACKET_DEBUG
    syslog(LOG_INFO, "SELEKT result %d\n", selectResult);
#endif
    if (selectResult < 0) {
      on_error("Thread socket select error", selectResult);
      break;
    }
    if (getStopIssued()) {
      syslog(LOG_INFO, "Stopping service loop (%u)\n", getpid());
      break;
    }
#ifdef PACKET_DEBUG
    // Report select status if we have pending write bytes.
    if (tcpBytes || unixBytes || 1) {
      syslog(LOG_INFO, "Selekt action r/w TCP %d %d/%d UNIX %d %d/%d (%u)\n",
	     unixBytes, 
	     FD_ISSET(tcpSock, &read_fd_set),
	     FD_ISSET(tcpSock, &write_fd_set),
	     tcpBytes,
	     FD_ISSET(unixSock, &read_fd_set),
	     FD_ISSET(unixSock, &write_fd_set),
	     getpid());
    }
#endif

    // First try to write any pending bytes.
    if (FD_ISSET(tcpSock, &write_fd_set) && unixBytes) {
#ifdef PACKET_DEBUG
      syslog(LOG_INFO, "TCP write %d (%u)\n", unixBytes, getpid());      
      hexdump("TCP Write", unixBuf, unixBytes);
#endif
      bytes = SSL_write(conn->ssl, unixBuf, unixBytes);
      if (bytes != unixBytes) {
        syslog(LOG_INFO, "Partial tcp write %d %d (%u)\n", bytes, unixBytes, getpid());
      }
      unixBytes = 0;
    }
    
    if (FD_ISSET(unixSock, &write_fd_set) && tcpBytes) {
#ifdef PACKET_DEBUG
      syslog(LOG_INFO, "UNIX write %d (%u)\n", tcpBytes, getpid());
      hexdump("UNIX Write", tcpBuf, tcpBytes);
#endif
      bytes = write(unixSock, tcpBuf, tcpBytes);
      if (bytes != tcpBytes) {
        syslog(LOG_INFO, "Partial unix write %d %d (%u)\n", bytes, tcpBytes, getpid());
      }
      tcpBytes = 0;
    }

    // Now read any new packets.
    if (FD_ISSET(tcpSock, &read_fd_set)) {
      if (tcpBytes) {
	on_error("Already have unwritten tcpBytes!", 0);
	continue;
      }
      // Read from the TCP socket.
      tcpBytes = SSL_read(conn->ssl, tcpBuf, BUFSIZE);
      if (tcpBytes < 0) {
	int err;
	err = SSL_get_error(conn->ssl, tcpBytes);
	// These errors are ok. We can get a select with no usable data.
	if ((err == SSL_ERROR_WANT_READ) || (err == SSL_ERROR_WANT_WRITE)) {
	  tcpBytes = 0;
	}
	syslog(LOG_INFO, "SSL error %d -> %d\n", err, tcpBytes);
      }
#ifdef PACKET_DEBUG
      syslog(LOG_INFO, "%s read %d (%u)\n", tcpname, tcpBytes, getpid());
      if (tcpBytes > 0) hexdump("TCP Read", tcpBuf, tcpBytes);
#endif
      if (tcpBytes < 0) {
	// Read error.
	on_error("Thread TCP socket read error", 0);
	break;
      }
      //continue;
    }

    if (FD_ISSET(unixSock, &read_fd_set)) {
      // Read from the UNIX socket.
      if (unixBytes) {
	on_error("Already have unwritten unixBytes!", 0);
	continue;
      }
      unixBytes = read(unixSock, unixBuf, BUFSIZE);
#ifdef PACKET_DEBUG
      syslog(LOG_INFO, "UNIX read %d (%u)\n", unixBytes, getpid());
      hexdump("UNIX Read", unixBuf, unixBytes);
#endif
      if (unixBytes < 0) {
	// Read error.
	on_error("Thread UNIX socket read error", 0);
	break;
      }
      //continue;
    }
  }
  syslog(LOG_INFO, "Free service resources loopcount %lld (%u)\n", loopcount, getpid());
  SSL_free(conn->ssl);
  close(tcpSock);
  SSL_CTX_free(conn->ctx);
  close(unixSock);
  close(fd);
  free(conn);
  serviceRunning = 0;
  stopService = 0;
  return NULL;
}

void InitSSL(void) {
  // Initialize SSL library stuff.
  SSL_library_init();
  ERR_load_crypto_strings();
  OpenSSL_add_all_algorithms();  /* Load cryptos, et.al. */
  SSL_load_error_strings();   /* Bring in and register error messages */
  syslog(LOG_INFO, "Version %s %lx\n", OPENSSL_VERSION_TEXT, OPENSSL_VERSION_NUMBER);
}
#endif // _USE_OPENSSL

// Information on the sockets used by nslink.
struct socketInfo {
  int sk;      // The communications socket (could be the same as net_sk or a unix socket).
  int net_sk;  // The TCP socket.
  int boxIndex;  // The hub number.
};


// Return socket connected to ip/port.  If 'secure' is true, create a unix
// socket to relay the data to and from an SSL TCP socket.

int connectsecure(struct sockaddr_storage *addr, int secure, struct socketInfo *skInfo, int nruns)
{
  int tcpSock;
  struct sockaddr_in *sockin;
  char addrstr[128];
#ifdef _USE_OPENSSL
  char unixname[256];
  struct secureConnection *conn;
  pthread_t tid;
  int clientSock, err;
  const SSL_METHOD *method;
#endif

  skInfo->sk = -1;
  skInfo->net_sk = -1;
  sockin = (struct sockaddr_in *)addr;
  format_inet_addr(addrstr, sizeof addrstr, addr);

  // Open TCP socket.
  syslog(LOG_INFO, "Open TCP socket %s port %d nruns %d (%u)\n", addrstr, ntohs(sockin->sin_port),
	 nruns, getpid());
  tcpSock = socket(addr->ss_family, SOCK_STREAM, 0);
  if (tcpSock < 0) {
    on_error("Could not create socket for NS-Link device", 0);
    return -1;
  }
  if (connect(tcpSock, (struct sockaddr *)addr, sizeof *addr)) {
    syslog(LOG_ERR,"Connect to %s failed", addrstr);
    on_error(NULL, 0);
    close(tcpSock);
    return -1;
  }
  int flag=1;
  setsockopt(tcpSock, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof flag);
  syslog(LOG_INFO, "TCP socket %d (%u)\n", tcpSock, getpid());
  if (!secure) {
    syslog(LOG_INFO, "Return TCP socket (%u)\n", getpid());
    skInfo->sk = tcpSock;
    skInfo->net_sk = (-1);
    // Return no error.
    return 0;
  }

#ifdef _USE_OPENSSL
  
  // Open SSL.
  syslog(LOG_INFO, "Open SSL socket (%u)\n", getpid());
  if (sslInited == 0) {
    InitSSL();
    sslInited = 1;
  }
  conn = malloc(sizeof(struct secureConnection));
  if (conn == NULL) {
    on_error("No memory for SSL connection", 0);
    return -1;
  }
  //method = TLSv1_2_client_method();  /* Create new client-method instance */
  //method = TLS_client_method();  /* Create new client-method instance */
#if defined(LWS_HAVE_TLS_CLIENT_METHOD)
  {
    syslog(LOG_INFO, "TLS_client_method (%u)\n", getpid());
    method = (SSL_METHOD *)TLS_client_method();
  }
#elif defined(LWS_HAVE_TLSV1_2_CLIENT_METHOD)
  {
    syslog(LOG_INFO, "TLSv1_2_client_method (%u)\n", getpid());
    method = (SSL_METHOD *)TLSv1_2_client_method();
  }
#else
  {
    // I think this is defined to be TLS_client_method these days.
    syslog(LOG_INFO, "SSLv23_client_method (%u)\n", getpid());
    method = (SSL_METHOD *)SSLv23_client_method();
  }
#endif

  conn->ctx = SSL_CTX_new(method);   /* Create new context */
  if (conn->ctx == NULL) {
    on_error("Failed to create SSL context", 0);
    return -1;
  }
  // Allow unsafe etc. Turn on the legacy option since it has been dropped from 3.0.3
  // which fails connection to the matrixssl server.
  //SSL_CTX_clear_options(conn->ctx, SSL_OP_LEGACY_SERVER_CONNECT);
  uint64_t options = SSL_CTX_get_options(conn->ctx);
  uint64_t change_option = SSL_OP_LEGACY_SERVER_CONNECT;
  syslog(LOG_INFO, "Current options %lx LEGACY_SERVER_CONNECT %lx\n", options, change_option);
  // This also works but may be more limited.
  //syslog(LOG_INFO, "Set option UNSAFE_LEGACY_RENEGOTIATION (%u)\n", getpid());
  //SSL_CTX_set_options(conn->ctx, SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION);
  syslog(LOG_INFO, "Set option LEGACY_SERVER_CONNECT (%u)\n", getpid());
  SSL_CTX_set_options(conn->ctx, change_option);
  options = SSL_CTX_get_options(conn->ctx);
  syslog(LOG_INFO, "New options %lx\n", options);
  
  conn->ssl = SSL_new(conn->ctx);
  conn->tcpSock = tcpSock;
  SSL_set_fd(conn->ssl, tcpSock);
  err = SSL_connect(conn->ssl);
  if (err < 0) {
    on_error("SSL_connect fail", err);
    return -1;
  }

#ifdef SOCKET_TEST
  // Start the socket test thread.
  shutSock = tcpSock;
  if (0) {
    err = pthread_create(&tidTest, NULL, &Tester, NULL);
    if (err != 0) {
      on_error("Can't create test thread", err);
    } else {
      syslog(LOG_INFO, "Test thread created successfully\n");
    }
  }
#endif

  // Create a unique name for the unix socket.
  //sprintf(unixname, "%s", addrstr);
  sprintf(unixname, "%s_%d", addrstr, nruns);
  syslog(LOG_INFO, "Create unix relay socket name %s (%u)\n", unixname, getpid());
  memset(&(conn->unixAddr), 0, sizeof(conn->unixAddr));
  conn->unixAddr.sun_family = AF_UNIX;
  *(conn->unixAddr.sun_path) = '\0';
  strncpy((conn->unixAddr.sun_path)+1, unixname, sizeof(conn->unixAddr.sun_path)-2);

  // Start the service thread.
  setStopService(0);
  err = pthread_create(&tid, NULL, &Service, (void *)conn);
  if (err != 0) {
    on_error("Can't create service thread", err);
    return -1;
  } else {
    syslog(LOG_INFO, "Service thread created successfully (%u)\n", getpid());
  }
  sleep(1);

  // Connect to the service thread unix socket.
  if ((clientSock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
    on_error("Unix socket create error", 0);
    return -1;
  }
  if (connect(clientSock, (struct sockaddr*)&(conn->unixAddr), sizeof(struct sockaddr_un)) == -1) {
    on_error("Unix socket connect error", 0);
    return -1;
  }
  syslog(LOG_INFO, "Successful client socket %d (%u)\n", clientSock, getpid());
  sleep(1);

  skInfo->sk = clientSock;
  skInfo->net_sk = tcpSock;
#endif // _USE_OPENSSL
  return 0;
}

#undef offsetof
#define offsetof(type, member)  __builtin_offsetof (type, member)


// Connect to specified address and send ID command.  If redboot is
// still running, disconnect, wait (with exponential backoff), and
// retry until NSLink app is found.  Return connected socket.
static int connect_tcp(struct sockaddr_storage *addr,  struct daemon_si_box_config *config, struct socketInfo *skInfo)
{
        struct id_request {
                uint8_t   cmd;
                uint8_t   index;
                uint8_t   lengthLSB;
                uint8_t   lengthMSB;
                uint32_t  addr;
        } __attribute__((packed));

        struct id_reply {
                uint32_t ipv4_addr;
                uint32_t ipv4_gate;
                uint32_t ipv4_serv;
                uint32_t ipv4_mask;
                uint32_t __unused0;
                uint32_t __unused1;
                char     vers_str[64];
                uint8_t  mac_addr[6];
                int32_t  model_id;
                uint8_t  arch_id;
                uint8_t  num_ports;
                uint16_t prog_id;
                uint8_t  board_rev;
                uint8_t  boot_major;
                uint8_t  boot_minor;
                uint8_t  num_apps;
                uint8_t  extended[1000];
        } __attribute__((packed));
	int nruns = 0;
        unsigned delay = 1;
        struct id_request idreq;
        struct id_reply   idrsp;
        char       addrstr[128];
	int err;

	sprintf(addrstr, "[%s]", inet_addr_string(addr));
	syslog(LOG_INFO,"Connecting to %s secure %d pid_file %s\n", addrstr, config->secure, pid_file);

        memset(&idreq,0,sizeof idreq);
        idreq.cmd = 5;
        idreq.index = 0;
        idreq.lengthLSB = (sizeof idreq);
        idreq.lengthMSB = 0;
        idreq.addr = 0;

        while (1) {
                int s,r;

                sleep(delay-1);
                if (delay < 32)
                        delay <<= 1;

#ifndef _USE_OPENSSL
		if (config->secure) {
		  syslog(LOG_ERR, "Cannot open a secure connection without OpenSSL development libraries installed");
		  syslog(LOG_ERR, "and the driver compiled with the SSL_FLAGS options.");
		  exit(-1);
		}
#endif
		// Force secure setting.
		//config->secure = 0;
		//syslog(LOG_INFO, "Connectsecure attempt %d\n", nruns);
		err = connectsecure(addr, config->secure, skInfo, nruns);
		nruns++;
		if (err < 0) {
		  syslog(LOG_ERR, "Connectsecure to %s %d failed", addrstr, err);
		  goto again;
		}
		  
		s = skInfo->sk;

		//syslog(LOG_INFO, "Wait for 30 seconds\n");
		//sleep(30);
                // send ID request
		syslog(LOG_INFO, "Write id request socket %d size %ld nruns %d (%u)\n", s,
		       sizeof idreq, nruns, getpid());
		if (nruns >= 0) {
		  r = write(s, &idreq, sizeof idreq);
		  //syslog(LOG_INFO, "Wrote %d\n", r);
		} else {
		  // Simulate some initial failures.
		  syslog(LOG_INFO, "Fake write %d\n", nruns);
		  r = 8;
		}
		    
                if (r < 0) {
		  syslog(LOG_ERR,"Write to %d %s failed: %m\n",s, addrstr);
		  goto again;
                }

                if (r < sizeof idreq) {
		  syslog(LOG_WARNING,"short write (%d) when sending ID command\n", r);
		  goto again;
                }

                // read response
                memset(&idrsp, 0, sizeof idrsp);

		//syslog(LOG_INFO, "Read id response socket %d size %ld\n", s, sizeof idrsp);
                r = read_with_timeout(s, &idrsp, sizeof idrsp, READ_TIMEOUT);
		//syslog(LOG_INFO, "Read %d bytes at least %ld\n", r, offsetof(struct id_reply, extended));

                if (r == 0) {
                        syslog(LOG_ERR,"Connection closed by hub\n");
                        goto again;
                }

                if (r < 0) {
		  syslog(LOG_ERR,"Error reading ID response: %s READ_TIMEOUT %d", strerror(errno), READ_TIMEOUT);
		  goto again;
                }

                if (r < offsetof(struct id_reply, extended)) {
		  // got a partial response, try another read
		  //syslog(LOG_INFO, "Read more!\n");
		  msleep(100);
		  int n = read_with_timeout(s, ((char *)&idrsp)+r, (sizeof idrsp)-r, READ_TIMEOUT);
		  if (n < 0) {
		    syslog(LOG_ERR,"Read from %s failed: %m\n",addrstr);
		    goto again;
		  }

		  if (r+n < offsetof(struct id_reply, extended)) {
		    syslog(LOG_ERR,"Error reading ID response: %s",strerror(errno));
		    goto again;
		  }
                }

                // got a response

                syslog(LOG_INFO,"TCP ID response from hub at %s: '%s' Model=%d Arch=%d NumPorts=%d ProgId=%d BoardRev=%d (%u)\n",
                       addrstr, idrsp.vers_str, ntohl(idrsp.model_id), idrsp.arch_id, idrsp.num_ports, ntohs(idrsp.prog_id), idrsp.board_rev, getpid());

                config->ModelId = ntohl(idrsp.model_id);
                config->arch_id = idrsp.arch_id;

                // if nslink application is running, we're done

                if (ntohs(idrsp.prog_id) == 1)
                        return s;

again:
#ifdef _USE_OPENSSL
		// We can't just go again if the unix socket opened.
		doStopService("connect_tcp");
#else
                close(s);	
#endif // _USE_OPENSSL
        }
}


static void setup_tcp_dev(int fd, struct daemon_si_box_config *config, int boxIndex)
{
        int sk, net_sk, rc;
	struct tcp_mode_params tcp_params;
	struct socketInfo skInfo;

        tcp_params.sk = -1;
        tcp_params.rbuf = NULL;
        tcp_params.rbuf_size = 0;
        tcp_params.wbuf = NULL;
        tcp_params.wbuf_size = 0;
	tcp_params.net_sk = -1;

        while (1) {
                // connect to hub and wait for nslink application
	        skInfo.boxIndex = boxIndex;
	        connect_tcp(&config->ip_addr, config, &skInfo);
		sk = skInfo.sk;
		net_sk = skInfo.net_sk;
                child_socket = net_sk;

                // hub should be up and running, turn the driver loose on it.

                tcp_params.boxid = boxIndex;
                tcp_params.sk = sk;
                tcp_params.rbuf = malloc(4096);
                tcp_params.rbuf_size = 4096;
                tcp_params.wbuf = malloc(1460);
                tcp_params.wbuf_size = 1460;
		tcp_params.net_sk = net_sk;
		
                if ((tcp_params.rbuf == NULL) || (tcp_params.wbuf == NULL)) {
                        syslog(LOG_ERR, "Could not allocate memory for NS-Link device, %s\n", inet_addr_string(&config->ip_addr));
                        exit(1);
                }

                syslog(LOG_INFO,"Child configuring hub (%u)\n", getpid());

                config_si(fd, config, boxIndex);

                syslog(LOG_INFO,"Child setting TCP sockets %d %d (%u)\n", sk, net_sk, getpid());
                rc = ioctl(fd, SI_SET_TCP, &tcp_params);
		syslog(LOG_INFO, "Return from SI_SET_TCP ioctl (%u)\n", getpid());
                if (rc) {
                        syslog(LOG_ERR, "Could not find driver for TCP hub %s\n", inet_addr_string(&config->ip_addr));
                        exit(1);
                }
		if (config->secure) {
#ifdef _USE_OPENSSL
		  doStopService("setup_tcp_dev");
#endif
		} else {
		  // Just close the tcp socket.
		  syslog(LOG_INFO, "Closing TCP socket child (%u)\n", getpid());
		  rc = close(child_socket);
		}
                free(tcp_params.wbuf);
                free(tcp_params.rbuf);
                sleep(1);
        }
}

static char *mactoa(unsigned char *m)
{
        static char buf[20];
        sprintf(buf,"%02x:%02x:%02x:%02x:%02x:%02x",m[0],m[1],m[2],m[3],m[4],m[5]);
        return buf;
}

typedef struct {
        unsigned char   destMac[6];
        unsigned char   srcMac[6];
        unsigned short  proto;
        unsigned char   prodId;
        unsigned char   concId;
        unsigned char   packetClass;
        unsigned short  length;
        unsigned char   adminCmd;
        unsigned char   adminData[1470];
}
__attribute__((packed))tAdminPacket;

#define AdminHeaderLen 20

typedef struct {
        unsigned char  tilde;
        unsigned short length;
        unsigned short cmd;
        uint32_t       offset;
        unsigned char  data[1460];
}
__attribute__((packed)) tMacBootPacket;

typedef struct {
        unsigned char  tilde;
        unsigned short length;
        unsigned short cmd;
        uint32_t       ioctl;
        uint32_t       startAddr;
        unsigned char checksum;
}
__attribute__((packed)) tMacGoPacket;

typedef struct {
        unsigned char    macAddr[6];
        unsigned char    flags;
        unsigned char    unused0;
        unsigned char    unused1;
        unsigned char    unused2;
        uint32_t         modelId;
        unsigned char    archId;
        unsigned char    numPorts;
        unsigned short   progId;
        unsigned char    boardRev;
        unsigned char    bootVersMajor;
        unsigned char    bootVersMinor;
        unsigned char    numAppRecords;
}
__attribute__((packed)) tMacIdResp;

typedef struct {
        unsigned char    macAddr[6];
        unsigned char    flags;
        unsigned char    unused0;
        unsigned char    unused1;
        unsigned char    unused2;
}
__attribute__((packed)) tOldMacIdResp;

#define ADMIN_BOOT           0
#define ADMIN_BOOT_PACKET    1
#define ADMIN_ID_QUERY       2
#define ADMIN_ID_REPLY       3
#define ADMIN_LOOP           4
#define ADMIN_RESET          5
#define ADMIN_TRACE          6
#define ADMIN_INIT           7
#define ADMIN_READMEM        8
#define ADMIN_WRITEMEM       9
#define ADMIN_TEST           10
#define ADMIN_DATA           11
#define ADMIN_ELBACK         12
#define ADMIN_READIO         13
#define ADMIN_WRITEIO        14
#define ADMIN_RSMODE         15
#define ADMIN_COMPMEM        16
#define ADMIN_MOVEMEM        17
#define ADMIN_TIMXMIT        18

#define ADMIN_IPINFO_SET    32
#define ADMIN_IPINFO_GET    33
#define ADMIN_IPINFO_ERASE  34

static unsigned short IntelOrderS(unsigned short v)
{
        unsigned char t;
        union {
                unsigned short s;
                unsigned char  b[2];
        } u;

        u.s = htons(v);
        t = u.b[1];
        u.b[1] = u.b[0];
        u.b[0] = t;
        return u.s;
}

// write MAC mode packet (padding to minimum Ethernet size if needed)
// WARNING: padding is added to caller's buffer, so it must always be
// at least 60 bytes long.
static int write_mac(int fd, void *buf, size_t count)
{
  if (count < 60)
    {
      memset(((char *)buf)+count, 0, 64-count);
      count = 60;
    }
  return write(fd, buf, count);
}

// returns: -1: failure
//           0: right application is running

static int connect_mac(unsigned char *destMac, int boxIndex, struct daemon_si_box_config *config)
{
        struct sockaddr_ll sockaddr;
        struct ifreq ifr;
        unsigned char myMac[6];
        tAdminPacket  cmd;
        tAdminPacket  rsp;
        tMacIdResp    *id =  (tMacIdResp*)&rsp.adminData;
        int n;
        char hubMacStr[20];
        int  sk;
        socklen_t u;
        int idRetryCount;

        strncpy(hubMacStr,mactoa(destMac),sizeof hubMacStr);

        //syslog(LOG_INFO, "Child %d creating packet socket\n", boxIndex);
        sk = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_COMTROL));

        if (sk < 0) {
                syslog(LOG_ERR, "Could not create socket for NS-Link hub %d at %s.",boxIndex,hubMacStr);
                exit(1);
        }

        // look up interface index (needed for bind())

        strncpy(ifr.ifr_name, config->interface, sizeof(ifr.ifr_name));
        ifr.ifr_name[(sizeof(ifr.ifr_name))-1] = '\0';

        if (ioctl(sk, SIOCGIFINDEX, &ifr)) {
                syslog(LOG_ERR,"Child %d couldn't find interface '%s'\n",boxIndex,config->interface);
                exit(1);
        }

        sockaddr.sll_family = AF_PACKET;
        sockaddr.sll_protocol = htons(ETH_P_COMTROL);
        sockaddr.sll_ifindex = ifr.ifr_ifindex;

        // only want to see packets from specified ethernet interface

        if (bind(sk,(struct sockaddr*) &sockaddr, sizeof sockaddr)) {
                syslog(LOG_ERR,"Child %d couldn't bind packet socket to %s\n",boxIndex,config->interface);
                exit(1);
        }

        child_socket = sk;

        u = sizeof sockaddr;
        getsockname(sk, (struct sockaddr*) &sockaddr, &u);

        memcpy(myMac,sockaddr.sll_addr,6);

        //syslog(LOG_INFO, "Child %d bound socket to %s (%s)\n", boxIndex, config->interface,mactoa(myMac));

        // stuff common to all commands we will send
        memcpy(cmd.destMac,destMac,6);
        memcpy(cmd.srcMac,myMac,6);
        cmd.proto = htons(ETH_P_COMTROL);
        cmd.prodId = 0xaa;
        cmd.concId = 0x80 + boxIndex;
        cmd.packetClass = 1;

        idRetryCount = 0;

        while (1) {
                // During DM Air development, it was found that the wireless
                // adapter/AP need to set up a bridging table before packets
                // will cross.  In TCP/IP this is done by arp/reponse.  In MAC
                // mode, we need to send a broadcast ADMIN_ID_QUERY without
                // requesting ownership.  This sets up the wireless components
                // with the target MAC addresses.

                memset(cmd.destMac, 0xFF, 6);
                cmd.adminCmd = ADMIN_ID_QUERY;
                memcpy(cmd.adminData,myMac,6);
                cmd.adminData[6] = 1;    // flags (bit0 == 0 ==> request ownership)
                cmd.length = IntelOrderS(8);
                write_mac(sk,&cmd,AdminHeaderLen+8);
                packetReadTimeout(sk, destMac, cmd.concId, &rsp, sizeof rsp, 200000);

                // Send ID request, request ownership
                memcpy(cmd.destMac,destMac,6);
                cmd.adminCmd = ADMIN_ID_QUERY;
                memcpy(cmd.adminData,myMac,6);
                cmd.adminData[6] = 0;    // flags (bit0 == 0 ==> request ownership)
                cmd.length = IntelOrderS(8);
                write_mac(sk,&cmd,AdminHeaderLen+8);

                if ((n=packetReadTimeout(sk, destMac, cmd.concId, &rsp, sizeof rsp, 200000)) <= 0) {
                        ++idRetryCount;
                        if ((idRetryCount < 5) || ((idRetryCount % 20) == 0))
                                syslog(LOG_ERR, "MAC ID request timeout for hub %d at %s",boxIndex,hubMacStr);
                        sleep(1);
                        if (idRetryCount > 5)
                                sleep(5);
                        continue;
                }

                idRetryCount = 0;

                if (n < (AdminHeaderLen + (int)sizeof(tOldMacIdResp)) || IntelOrderS(rsp.length) < 7) {
                        syslog(LOG_ERR, "MAC ID response too short from hub %d at %s",boxIndex,hubMacStr);
                        sleep(5);
                        continue;
                }

                if (rsp.adminCmd != ADMIN_ID_REPLY) {
                        syslog(LOG_ERR, "MAC ID response has wrong cmd type from hub %d at %s",boxIndex,hubMacStr);
                        sleep(5);
                        continue;
                }

                id->progId = ntohs(id->progId);
                id->modelId = ntohl(id->modelId);

                if (n >= (int)sizeof(tMacIdResp) &&
                    IntelOrderS(rsp.length) >= 22 &&
                    id->macAddr[3] != 0x06) {
                        // find model/version string in extended response data
                        unsigned char *p = &id->numAppRecords + 1 + 4*id->numAppRecords;     // start of extended data
                        unsigned char *end = (unsigned char*)&rsp + IntelOrderS(rsp.length); // end of extended data
                        char *version = "";

                        while (p<end) {
                                unsigned tag = *p++;
                                unsigned len = *p++;
                                if (tag == 1 && len > 0) {
                                        version = (char*)p;
                                        break;
                                }
                                p += len;
                        }

                        syslog(LOG_INFO, "MAC ID response from hub %d at %s '%s' flags=%02x, Model=%u Arch=%d NumPorts=%d ProgId=%d BoardRev=%d",
                               boxIndex, hubMacStr, version, id->flags, (unsigned int)id->modelId, id->archId, id->numPorts, id->progId, id->boardRev);
                } else {
                        // must be old box!
                        syslog(LOG_INFO, "MAC ID response from hub %d at %s flags=%02x\n", boxIndex,hubMacStr,id->flags);

                        // we have to look at addr and flag values to figure out who we're talking to :(
                        // kdr  Bad Bad Bad...  Magic numbers....
                        if ((id->macAddr[3] == 0x10 || id->macAddr[3] == 0x11))
                                id->archId = 2;
                        else if (id->macAddr[3] == 0x06)
                                id->archId = 1;
                        else
                                id->archId = 0;         // don't know

                        if (id->flags & 0x01)
                                id->progId = 1;
                        else
                                id->progId = 0;
                }

                config->arch_id = id->archId;
                config->ModelId = id->modelId;

                // right app running?
                if (id->progId == 1) {
                        close(sk);
                        return 0;
                }

                // wait for nslink app
                sleep(5);
        }
}

static void setup_mac_dev(int fd,   struct daemon_si_box_config *config, int boxIndex)
{
        int rc;
        struct mac_mode_params mac_params;

        while (1) {
                // download executable binary file to hub and restart it if needed

                if (connect_mac(config->ether_addr,boxIndex,config)) {
                        syslog(LOG_ERR,"Couldn't detect NS-Link device, %s.\n", mactoa(config->ether_addr));
                        sleep(5);
                        continue;
                }

                // hub should be up and running, turn the driver loose on it.
                //syslog(LOG_INFO,"Child %d configuring hub\n",boxIndex);

                rc = config_si(fd, config, boxIndex);

                if (rc && (rc != EBUSY)) {
                        syslog(LOG_ERR, "Could not configure unit...exiting!\n");
                        exit(1);
                }

                //syslog(LOG_INFO,"Child %d setting Mac\n",boxIndex);

                mac_params.boxid = boxIndex;
		// If all goes well, no return from here.
                rc = ioctl(fd, SI_SET_MAC, &mac_params);

                if (rc) {
                        syslog(LOG_ERR, "Could not set device, %s.\n", mactoa(config->ether_addr));
                        exit(1);
                }
        }
}


static void setup_devices(int fd)
{
        int i;
        int newpid;
        struct daemon_si_box_config *config;
	
        for (i=0; i< num_box_config; ++i) {
                config = box_config + i;

                newpid = fork();

                if (newpid) {
                        /* parent */
                        child_pids[i] = newpid;
                        continue;
                }

                /* child */
		driver_fd = fd;
                if (config->ip_addr.ss_family) {
		  // Run explicit shutdown in TCP/SSL mode.
		  struct sigaction action;
		  memset(&action, 0, sizeof(action));
		  action.sa_handler = sigterm_handler;
		  sigaction(SIGTERM, &action, NULL);
		  sigaction(SIGINT, &action, NULL);
		  syslog(LOG_INFO, "Install sigterm_handler (%u)\n", getpid());
		  //signal(SIGTERM, sigterm_handler);
		  //signal(SIGINT, sigterm_handler);
		  setup_tcp_dev(fd,config,i);
		}
                else {
		  setup_mac_dev(fd,config,i);
		}
        }//end for
}


int main(int argc, char **argv)
{
        int major, fd;

        if (getuid()) {
                fprintf(stderr, "You must be root to run NS-Link daemon!\n");
                exit(1);
        }

        openlog("nslinkd", LOG_NDELAY | LOG_PID, LOG_DAEMON);
        syslog(LOG_INFO, "Starting NS-Link daemon, version %s (%s)",SI_VERSION, SI_DATE);

        parse_config("/etc/nslink.conf");

        major = get_major_number("NSLinkctl");

        if (major < 0) {
                syslog(LOG_ERR, "Couldn't load control file");
                fprintf(stderr, "Couldn't find control file.  %s\n", not_loaded);
                exit(1);
        }

        fd = open_dev(major << 8, O_RDWR);
        if (fd < 0) {
                syslog(LOG_ERR, "Couldn't open control device: %s", strerror(errno));
                fprintf(stderr, "Couldn't open control device.  %s\n", not_loaded);
                perror("  open_dev()");
                exit(1);
        }

        check_pid_file();
	driver_fd = fd;

        if (become_daemon(fd) < 0) {
                syslog(LOG_ERR, "Couldn't become daemon");
                fprintf(stderr, "Error forking child process while "
                        "starting daemon mode\n");
                exit(1);
        }

        put_pid_file();
        signal(SIGHUP, sighup_handler);
	syslog(LOG_INFO, "Install sigterm handler (%u)\n", getpid());
        signal(SIGTERM, sigterm_handler);
        signal(SIGINT, sigterm_handler);
        create_devices(fd);
        setup_devices(fd);
        // JS - Has to add this delay otherwise the parent process of nslinkd
        // daemon gets killed at bootup on Ubuntu system. Don't know why though.
        sleep(3);
        pause();
        return 0;
}
